mod download;
mod state;
mod upload;
mod util;
mod zip;

use std::{fs::File, path::PathBuf};

use actix_web::{
    get, middleware::Logger, web, App, HttpRequest, HttpResponse, HttpServer, Responder,
};
use actix_web_actors::ws;
use serde::{Deserialize, Serialize};
use state::PersistentState;
use time::OffsetDateTime;
use tokio::sync::RwLock;

const APP_NAME: &str = "transbeam";

pub struct UploadedFile {
    name: String,
    size: u64,
    modtime: OffsetDateTime,
}

impl UploadedFile {
    fn new(name: &str, size: u64, modtime: OffsetDateTime) -> Self {
        Self {
            name: sanitise_file_name::sanitise(name),
            size,
            modtime,
        }
    }
}

#[derive(Clone, Deserialize, Serialize)]
pub struct DownloadableFile {
    name: String,
    size: u64,
    #[serde(with = "state::timestamp")]
    modtime: OffsetDateTime,
}

type AppData = web::Data<RwLock<PersistentState>>;

fn storage_dir() -> PathBuf {
    PathBuf::from(std::env::var("STORAGE_DIR").unwrap_or_else(|_| String::from("storage")))
}

#[get("/download/{file_code}")]
async fn handle_download(
    req: HttpRequest,
    path: web::Path<String>,
    data: AppData,
) -> actix_web::Result<HttpResponse> {
    let file_code = path.into_inner();
    if !util::is_ascii_alphanumeric(&file_code) {
        return Ok(HttpResponse::NotFound().finish());
    }
    let data = data.read().await;
    let info = data.lookup_file(&file_code);
    if let Some(info) = info {
        let storage_path = storage_dir().join(file_code);
        let file = File::open(&storage_path)?;
        Ok(download::DownloadingFile {
            file,
            storage_path,
            info,
        }
        .into_response(&req))
    } else {
        Ok(HttpResponse::NotFound().finish())
    }
}

#[get("/upload")]
async fn handle_upload(req: HttpRequest, stream: web::Payload, data: AppData) -> impl Responder {
    ws::start(upload::Uploader::new(data), &req, stream)
}

#[actix_web::main]
async fn main() -> std::io::Result<()> {
    env_logger::init();

    let data: AppData = web::Data::new(RwLock::new(PersistentState::load().await?));

    let static_dir =
        PathBuf::from(std::env::var("STATIC_DIR").unwrap_or_else(|_| String::from("static")));
    let port = std::env::var("PORT")
        .ok()
        .and_then(|p| p.parse::<u16>().ok())
        .unwrap_or(8080);

    HttpServer::new(move || {
        App::new()
            .app_data(data.clone())
            .wrap(Logger::default())
            .service(handle_upload)
            .service(handle_download)
            .service(actix_files::Files::new("/", static_dir.clone()).index_file("index.html"))
    })
    .bind(("127.0.0.1", port))?
    .run()
    .await?;
    Ok(())
}
